@@ -9,6 +9,7 @@ import (
 	"time"
 
 	"github.com/google/wire"
+	"github.com/samber/lo"
 	"golang.org/x/exp/slices"
 	"golang.org/x/xerrors"
 
@@ -96,19 +97,33 @@ func (s Scanner) Scan(ctx context.Context, target, artifactKey string, blobKeys
 	}
 
 	var eosl bool
-	var results types.Results
+	var results, pkgResults types.Results
 
-	// Scan OS packages and language-specific dependencies
+	// Fill OS packages and language-specific packages
+	if options.ListAllPackages {
+		if res := s.osPkgsToResult(target, artifactDetail, options); res != nil {
+			pkgResults = append(pkgResults, *res)
+		}
+		pkgResults = append(pkgResults, s.langPkgsToResult(artifactDetail)...)
+	}
+
+	// Scan packages for vulnerabilities
 	if slices.Contains(options.SecurityChecks, types.SecurityCheckVulnerability) {
 		var vulnResults types.Results
-		vulnResults, eosl, err = s.checkVulnerabilities(target, artifactDetail, options)
+		vulnResults, eosl, err = s.scanVulnerabilities(target, artifactDetail, options)
 		if err != nil {
 			return nil, nil, xerrors.Errorf("failed to detect vulnerabilities: %w", err)
 		}
 		if artifactDetail.OS != nil {
 			artifactDetail.OS.Eosl = eosl
 		}
+		// Merge package results into vulnerability results
+		s.fillPkgsInVulns(pkgResults, vulnResults)
+
 		results = append(results, vulnResults...)
+	} else {
+		// If vulnerability scanning is not enabled, it just adds package results.
+		results = append(results, pkgResults...)
 	}
 
 	// Scan IaC config files
@@ -123,6 +138,12 @@ func (s Scanner) Scan(ctx context.Context, target, artifactKey string, blobKeys
 		results = append(results, secretResults...)
 	}
 
+	// Scan licenses
+	if slices.Contains(options.SecurityChecks, types.SecurityCheckLicense) {
+		licenseResults := s.scanLicenses(artifactDetail, options.LicenseCategories)
+		results = append(results, licenseResults...)
+	}
+
 	// For WASM plugins and custom analyzers
 	if len(artifactDetail.CustomResources) != 0 {
 		results = append(results, types.Result{
@@ -131,12 +152,6 @@ func (s Scanner) Scan(ctx context.Context, target, artifactKey string, blobKeys
 		})
 	}
 
-	// Scan licenses
-	if slices.Contains(options.SecurityChecks, types.SecurityCheckLicense) {
-		licenseResults := s.scanLicenses(artifactDetail, options.LicenseCategories)
-		results = append(results, licenseResults...)
-	}
-
 	for i := range results {
 		// Fill vulnerability details
 		s.vulnClient.FillInfo(results[i].Vulnerabilities)
@@ -151,11 +166,51 @@ func (s Scanner) Scan(ctx context.Context, target, artifactKey string, blobKeys
 	return results, artifactDetail.OS, nil
 }
 
-func shouldScanMisconfig(securityChecks []string) bool {
-	return slices.Contains(securityChecks, types.SecurityCheckConfig) || slices.Contains(securityChecks, types.SecurityCheckRbac)
+func (s Scanner) osPkgsToResult(target string, detail ftypes.ArtifactDetail, options types.ScanOptions) *types.Result {
+	if len(detail.Packages) == 0 {
+		return nil
+	}
+	if detail.OS != nil {
+		target = fmt.Sprintf("%s (%s %s)", target, detail.OS.Family, detail.OS.Name)
+	}
+	pkgs := detail.Packages
+	if options.ScanRemovedPackages {
+		pkgs = mergePkgs(pkgs, detail.HistoryPackages)
+	}
+	sort.Slice(pkgs, func(i, j int) bool {
+		return strings.Compare(pkgs[i].Name, pkgs[j].Name) <= 0
+	})
+	return &types.Result{
+		Target:   target,
+		Class:    types.ClassOSPkg,
+		Type:     detail.OS.Family,
+		Packages: pkgs,
+	}
 }
 
-func (s Scanner) checkVulnerabilities(target string, detail ftypes.ArtifactDetail, options types.ScanOptions) (
+func (s Scanner) langPkgsToResult(detail ftypes.ArtifactDetail) types.Results {
+	var results types.Results
+	for _, app := range detail.Applications {
+		if len(app.Libraries) == 0 {
+			continue
+		}
+		target := app.FilePath
+		if t, ok := pkgTargets[app.Type]; ok && target == "" {
+			// When the file path is empty, we will overwrite it with the pre-defined value.
+			target = t
+		}
+
+		results = append(results, types.Result{
+			Target:   target,
+			Class:    types.ClassLangPkg,
+			Type:     app.Type,
+			Packages: app.Libraries,
+		})
+	}
+	return results
+}
+
+func (s Scanner) scanVulnerabilities(target string, detail ftypes.ArtifactDetail, options types.ScanOptions) (
 	types.Results, bool, error) {
 	var eosl bool
 	var results types.Results
@@ -171,7 +226,7 @@ func (s Scanner) checkVulnerabilities(target string, detail ftypes.ArtifactDetai
 	}
 
 	if slices.Contains(options.VulnType, types.VulnTypeLibrary) {
-		libResults, err := s.scanLibrary(detail.Applications, options)
+		libResults, err := s.scanLangPkgs(detail.Applications)
 		if err != nil {
 			return nil, false, xerrors.Errorf("failed to scan application libraries: %w", err)
 		}
@@ -183,7 +238,7 @@ func (s Scanner) checkVulnerabilities(target string, detail ftypes.ArtifactDetai
 
 func (s Scanner) scanOSPkgs(target string, detail ftypes.ArtifactDetail, options types.ScanOptions) (
 	*types.Result, bool, error) {
-	if detail.OS == nil {
+	if detail.OS == nil || detail.OS.Family == "" {
 		log.Logger.Debug("Detected OS: unknown")
 		return nil, false, nil
 	}
@@ -194,45 +249,24 @@ func (s Scanner) scanOSPkgs(target string, detail ftypes.ArtifactDetail, options
 		pkgs = mergePkgs(pkgs, detail.HistoryPackages)
 	}
 
-	result, eosl, err := s.detectVulnsInOSPkgs(target, detail.OS.Family, detail.OS.Name, detail.Repository, pkgs)
-	if err != nil {
-		return nil, false, xerrors.Errorf("failed to scan OS packages: %w", err)
-	} else if result == nil {
-		return nil, eosl, nil
-	}
-
-	if options.ListAllPackages {
-		sort.Slice(pkgs, func(i, j int) bool {
-			return strings.Compare(pkgs[i].Name, pkgs[j].Name) <= 0
-		})
-		result.Packages = pkgs
-	}
-
-	return result, eosl, nil
-}
-
-func (s Scanner) detectVulnsInOSPkgs(target, osFamily, osName string, repo *ftypes.Repository, pkgs []ftypes.Package) (*types.Result, bool, error) {
-	if osFamily == "" {
-		return nil, false, nil
-	}
-	vulns, eosl, err := s.ospkgDetector.Detect("", osFamily, osName, repo, time.Time{}, pkgs)
+	vulns, eosl, err := s.ospkgDetector.Detect("", detail.OS.Family, detail.OS.Name, detail.Repository, time.Time{}, pkgs)
 	if err == ospkgDetector.ErrUnsupportedOS {
 		return nil, false, nil
 	} else if err != nil {
 		return nil, false, xerrors.Errorf("failed vulnerability detection of OS packages: %w", err)
 	}
 
-	artifactDetail := fmt.Sprintf("%s (%s %s)", target, osFamily, osName)
+	artifactDetail := fmt.Sprintf("%s (%s %s)", target, detail.OS.Family, detail.OS.Name)
 	result := &types.Result{
 		Target:          artifactDetail,
 		Vulnerabilities: vulns,
 		Class:           types.ClassOSPkg,
-		Type:            osFamily,
+		Type:            detail.OS.Family,
 	}
 	return result, eosl, nil
 }
 
-func (s Scanner) scanLibrary(apps []ftypes.Application, options types.ScanOptions) (types.Results, error) {
+func (s Scanner) scanLangPkgs(apps []ftypes.Application) (types.Results, error) {
 	log.Logger.Infof("Number of language-specific files: %d", len(apps))
 	if len(apps) == 0 {
 		return nil, nil
@@ -255,6 +289,8 @@ func (s Scanner) scanLibrary(apps []ftypes.Application, options types.ScanOption
 		vulns, err := library.Detect(app.Type, app.Libraries)
 		if err != nil {
 			return nil, xerrors.Errorf("failed vulnerability detection of libraries: %w", err)
+		} else if len(vulns) == 0 {
+			continue
 		}
 
 		target := app.FilePath
@@ -263,16 +299,12 @@ func (s Scanner) scanLibrary(apps []ftypes.Application, options types.ScanOption
 			target = t
 		}
 
-		libReport := types.Result{
+		results = append(results, types.Result{
 			Target:          target,
 			Vulnerabilities: vulns,
 			Class:           types.ClassLangPkg,
 			Type:            app.Type,
-		}
-		if options.ListAllPackages {
-			libReport.Packages = app.Libraries
-		}
-		results = append(results, libReport)
+		})
 	}
 	sort.Slice(results, func(i, j int) bool {
 		return results[i].Target < results[j].Target
@@ -280,6 +312,17 @@ func (s Scanner) scanLibrary(apps []ftypes.Application, options types.ScanOption
 	return results, nil
 }
 
+func (s Scanner) fillPkgsInVulns(pkgResults, vulnResults types.Results) {
+	// Fill vulnerability results in package results
+	for i := range vulnResults {
+		if r, found := lo.Find(pkgResults, func(r types.Result) bool {
+			return r.Class == vulnResults[i].Class && r.Target == vulnResults[i].Target
+		}); found {
+			vulnResults[i].Packages = r.Packages
+		}
+	}
+}
+
 func (s Scanner) misconfsToResults(misconfs []ftypes.Misconfiguration) types.Results {
 	log.Logger.Infof("Detected config files: %d", len(misconfs))
 	var results types.Results
@@ -475,3 +518,7 @@ func mergePkgs(pkgs, pkgsFromCommands []ftypes.Package) []ftypes.Package {
 	}
 	return pkgs
 }
+
+func shouldScanMisconfig(securityChecks []string) bool {
+	return slices.Contains(securityChecks, types.SecurityCheckConfig) || slices.Contains(securityChecks, types.SecurityCheckRbac)
+}
